"""Contain utility functions for icclim."""

from __future__ import annotations

from datetime import datetime

import dateparser

from icclim.exception import InvalidIcclimArgumentError


def read_date(in_date: str | datetime) -> datetime:
    """
    Read a date from a string or return the date if it is already a datetime object.

    Parameters
    ----------
    in_date: str | datetime
        A string representing a date or a datetime object.

    Returns
    -------
    datetime
        A datetime object.
    """
    if isinstance(in_date, datetime):
        return in_date
    date = dateparser.parse(in_date)
    if date is None:
        msg = (
            f"The date {in_date} does not have a valid format."
            " You can use various formats such as '2 december', '02-12',"
            " '1994-12-02'..."
        )
        raise InvalidIcclimArgumentError(msg)
    return date


def is_number_sequence(values: object) -> bool:
    """Return True if values is a sequence of numbers."""
    return isinstance(values, (tuple, list)) and all(
        (isinstance(x, (float, int)) for x in values),
    )



"""
Contain the CfCalendar class and the CfCalendarRegistry class.

These classes are used to represent and manage the CF calendars used in the
climate indices calculation.
"""

from __future__ import annotations

import dataclasses
from collections.abc import Callable

import numpy as np
import xarray as xr
from xarray import DataArray

from icclim._core.model.registry import Registry


@dataclasses.dataclass
class CfCalendar:
    """
    Represents a CF calendar.

    Parameters
    ----------
    aliases : list[str]
        All the possible aliases or poorly typed calendar names targeting the same
        calendar.
    is_leap : Callable
        It expects a DataArray argument of years such as `da.time.dt.year`.
        Returns a mask of the input telling if the value is part of a leap year or not.

    Attributes
    ----------
    aliases : list[str]
        All the possible aliases or poorly typed calendar names targeting the same
        calendar.
    is_leap : Callable
        It expects a DataArray argument of years such as `da.time.dt.year`.
        Returns a mask of the input telling if the value is part of a leap year or not.

    Methods
    -------
    name() -> str:
        Returns the name of the calendar.

    """

    aliases: list[str]
    is_leap: Callable[[DataArray], np.ndarray]

    @property
    def name(self) -> str:
        """
        Returns the name of the calendar.

        Returns
        -------
        str
            The name of the calendar.
        """
        return self.aliases[0]


# TODO @bzah: the whole class might be useless with the latest cftime
#      (we don't need our own CfCalendar if we can do `da.time.dt.is_leap_year`)
# https://github.com/cerfacs-globc/icclim/issues/289
class CfCalendarRegistry(Registry[CfCalendar]):
    """
    Calendars known in CF plus some additional custom aliases for convenience.

    Parameters
    ----------
    aliases : list[str]
        All the possible aliases or poorly typed calendar names targeting the same
        calendar.
    is_leap : Callable
        It expects a DataArray argument of years such as `da.time.dt.year`.
        Returns a mask of the input telling if the value is part of a leap year or not.

    Attributes
    ----------
    _item_class : type
        The class of the items in the registry.

    Methods
    -------
    get_item_aliases(item: CfCalendar) -> list[str]:
        Get the aliases of a CfCalendar item.

    """

    _item_class = CfCalendar

    NO_LEAP = CfCalendar(
        ["noleap", "no_leap", "days_365", "days365", "365_day", "365day"],
        lambda da: np.full_like(da.shape, False, dtype=bool),
    )
    DAYS_360 = CfCalendar(
        ["360_day", "days_360", "360day", "days360"],
        lambda da: np.full_like(da.shape, False, dtype=bool),
    )
    ALL_LEAP = CfCalendar(
        ["all_leap", "allleap", "days_366", "days366", "366_day", "366day"],
        lambda da: np.full_like(da.shape, True, dtype=bool),
    )
    PROLEPTIC_GREGORIAN = CfCalendar(
        ["proleptic_gregorian", "prolepticgregorian"],
        lambda da: _proleptic_gregorian_leap(da).values,
    )
    JULIAN = CfCalendar(["julian"], lambda da: _julian_leap(da).values)
    STANDARD = CfCalendar(
        ["standard", "gregorian"],
        lambda da: _standard_leap(da).values,
    )
    # Not sure what to do with none calendar
    NONE = CfCalendar(["none"], lambda da: _standard_leap(da).values)

    @staticmethod
    def get_item_aliases(item: CfCalendar) -> list[str]:
        """
        Get the aliases of a CfCalendar item.

        Parameters
        ----------
        item : CfCalendar
            The CfCalendar item.

        Returns
        -------
        list[str]
            The aliases of the CfCalendar item.
        """
        return list(map(str.upper, item.aliases))


def _proleptic_gregorian_leap(years: DataArray) -> DataArray:
    """
    Calculate if the years are part of a leap year in the proleptic Gregorian calendar.

    Parameters
    ----------
    years : DataArray
        The years to check.

    Returns
    -------
    DataArray
        A boolean array indicating if the years are part of a leap year.
    """
    return np.logical_or(
        years % 400 == 0,
        np.logical_and(years % 100 != 0, years % 4 == 0),
    )


def _julian_leap(years: DataArray) -> DataArray:
    """
    Calculate if the years are part of a leap year in the Julian calendar.

    Parameters
    ----------
    years : DataArray
        The years to check.

    Returns
    -------
    DataArray
        A boolean array indicating if the years are part of a leap year.
    """
    return years % 4 == 0


def _standard_leap(years: DataArray) -> DataArray:
    """
    Calculate if the years are part of a leap year in the standard Gregorian calendar.

    Parameters
    ----------
    years : DataArray
        The years to check.

    Returns
    -------
    DataArray
        A boolean array indicating if the years are part of a leap year.
    """
    res = xr.full_like(years, False)
    res[years < 1582] = _julian_leap(years[years < 1582])
    res[years >= 1582] = _proleptic_gregorian_leap(years[years >= 1582])
    return res



"""Contain the StandardIndex data class."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from xarray import DataArray

    from icclim._core.model.index_group import IndexGroup
    from icclim._core.model.indicator import Indicator
    from icclim._core.model.standard_variable import StandardVariable
    from icclim._core.model.threshold import Threshold


@dataclass
class StandardIndex:
    """
    Standard Index data class.

    It is used to describe how a GenericIndicator should be setup to compute a climate
    index that has been defined in the literature (such as ECA&D's ATBD document).


    Attributes
    ----------
    short_name: str
        The index name used in the output.
    compute: Callable
        The function to compute the index. It usually wraps a xclim functions.
    group: IndexGroup
        The index group category.
    variables: List[List[str]]
        The Cf variables needed to compute the index.
        The variable are individually described by a list of aliases.
    qualifiers: List[str] | None
        ``optional`` List of configuration to compute the index.
        Used internally to generate modules for C3S.
    source: str | None
        Where the index definition comes from.
    definition: str | None
        A formal definition of the index. It should describe what kind of output
        the user is expected to obtain.
    """

    short_name: str
    group: IndexGroup
    input_variables: list[StandardVariable] | None  # None when index is generic
    indicator: Indicator
    # TODO @bzah: merge qualifiers with group into a Set of qualifiers ?
    # https://github.com/cerfacs-globc/icclim/issues/289
    reference: str
    qualifiers: list[str] | None = None
    source: str | None = None
    definition: str | None = None
    threshold: str | None | Threshold | Sequence[str | Threshold] = None
    output_unit: str | None = None
    # additional, index specific args
    rolling_window_width: int | None = None
    doy_window_width: int | None = None
    min_spell_length: int | None = None

    def __str__(self) -> str:
        """Return a string representation of the index."""
        return f"{self.group} | {self.short_name} | {self.definition}"

    def __call__(self, *args, **kwargs) -> DataArray:
        """Call the indicator's compute method."""
        self.indicator(*args, **kwargs)

    def __eq__(self, other: object) -> bool:
        """Return True if the two StandardIndex are equal."""
        if not isinstance(other, StandardIndex):
            return False
        return (
            self.indicator == other.indicator
            and self.threshold == other.threshold
            and self.output_unit == other.output_unit
            and self.rolling_window_width == other.rolling_window_width
            and self.doy_window_width == other.doy_window_width
            and self.min_spell_length == other.min_spell_length
        )

    def clone(self) -> StandardIndex:
        """Return a deep copy of the index."""
        from copy import deepcopy

        return deepcopy(self)



"""
Contain the IndexConfig class.

It holds the compiled configuration for the computation of climate indices.
"""

from __future__ import annotations

import dataclasses
from collections.abc import Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from icclim._core.climate_variable import ClimateVariable
    from icclim._core.model.indicator import Indicator
    from icclim._core.model.logical_link import LogicalLink
    from icclim._core.model.netcdf_version import NetcdfVersion
    from icclim._core.model.quantile_interpolation import QuantileInterpolation
    from icclim.frequency import Frequency


@dataclasses.dataclass
class IndexConfig:
    """
    Configuration class for defining climate index parameters.

    Parameters
    ----------
    frequency : Frequency
        The time frequency of the output. Built from ``slice_mode``.
    climate_variables : list[ClimateVariable]
        The list of climate variables used in the index calculation.
    min_spell_length : int | None
        The minimum spell length for the index calculation.
        None if the index is not a spell index.
    rolling_window_width : int | None
        The width of the rolling window for the index calculation.
        None if the index is not a rolling index.
    out_unit : str | None
        The output unit for the index calculation.
        Optional, used to override the default unit.
    callback : Callable[[int], None] | None
        The callback function for progress updates during the index calculation.
        Deprecated.
    netcdf_version : NetcdfVersion
        The version of the NetCDF file format to use for saving the index results.
        Default is NetcdfVersion.NETCDF4.
    save_thresholds : bool
        Flag indicating whether to save the threshold values used in the index
        calculation.
    interpolation : QuantileInterpolation
        The interpolation method to use for calculating quantiles/percentiles.
    is_compared_to_reference : bool
        Flag indicating whether the index is compared to a reference period.
    reference_period : tuple[str, str] | None
        The reference period for the index calculation.
    indicator_name : str
        The name of the index.
    logical_link : LogicalLink
        The logical link to use for combining multiple indices.
    coef : float | None
        The coefficient to apply to the index values.
    date_event : bool
        Flag indicating whether the index represents a date or an event.
    sampling_method : str
        The sampling method to use for the index calculation.
        In conjonction with the Frequency, it is used on specific indices such as the
        anomaly (a.k.a diff_of_means) to determine if the reference period and the
        studied period should be grouped by or resampled.
        It can be either 'group_by', 'resample', or
        'group_by_ref_and_resample_study'.
        'group_by' will group the data by the specified frequency, for example every
        data of every January together.
        'resample' will resample the data to the specified frequency, for example every
        days of each month independently together.
        'group_by_ref_and_resample_study' will group the reference data by the specified
        frequency and resample the study data to the same frequency.
        This last method allows for example to compare each January, independently, of
        the study period to every January of the reference period.
        This is typically used to compare the each month of the studied period
        to a normal (the reference) of many aggregated years.
    rename : str | None
        The new name for the output variable.
        Optional, used to override the default index name.
    indicator : Indicator
        The indicator to be computed.
    reference : str
        The reference value for the index calculation.
    """

    frequency: Frequency
    climate_variables: list[ClimateVariable]
    min_spell_length: int | None
    rolling_window_width: int | None
    out_unit: str | None
    callback: Callable[[int], None] | None
    netcdf_version: NetcdfVersion
    save_thresholds: bool
    interpolation: QuantileInterpolation
    is_compared_to_reference: bool
    reference_period: tuple[str, str] | None
    indicator_name: str
    logical_link: LogicalLink
    coef: float | None
    date_event: bool
    sampling_method: str
    rename: str | None
    indicator: Indicator
    reference: str



"""Contains the InFileDictionary class."""

from __future__ import annotations

from typing import TYPE_CHECKING

# for Python <3.11 with (Not)Required
from typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    from icclim._core.model.icclim_types import InFileBaseType
    from icclim._core.model.threshold import Threshold


class InFileDictionary(TypedDict):
    """
    Dictionary grouping in_files and var_name functionnalities.

    Attributes
    ----------
    study : InFileBaseType
        Study input file.
    thresholds : NotRequired[Threshold | None]
        Thresholds to apply to the study input file.

    Examples
    --------
    .. code-block:: python

        threshold = build_threshold(operator=">", value=["per-1.nc", "per-2.nc"])
        in_files = {
            "tasmax": {
                "study": "tasmax-store.zarr",
                "threshold": threshold,
            },
            "pr": "pr.nc",
            "tasmin": {"study": "tasmin.nc"},
        }

    Notes
    -----
    It also allows to use a different input for thresholds such as percentiles.
    """

    study: InFileBaseType
    thresholds: NotRequired[Threshold | None]



"""Contain the Registry class, a fancy enum replacement."""

from __future__ import annotations

from abc import ABC
from copy import deepcopy
from typing import Generic, TypeVar

from icclim.exception import InvalidIcclimArgumentError

T = TypeVar("T")


class Registry(Generic[T], ABC):
    """
    Registry classes acts as fancy enums.

    It allows to easily store and find constants of similar type.
    Registries are namespaces, so there is no need to instantiate it or any of
    its subclasses, every item is a class attribute.

    Notes
    -----
    Registries are not meant to store large collections, they are just fancy lookup
    tables for items with aliases and no case sensitivity.
    """

    _item_class: type  # runtime type for the generic `T`

    @classmethod
    def lookup(cls, query: T | str) -> T:
        """
        Look up an item in the registry.

        Parameters
        ----------
        query : T or str
            The item to look up. It can be either an instance of the item class or a
            string.

        Returns
        -------
        T
            The found item.

        Raises
        ------
        InvalidIcclimArgumentError
            If the item is not found in the registry.

        Notes
        -----
        This method performs a case-insensitive lookup.
        It first checks if the query is an instance of the item class, and if so,
        returns a deep copy of the query.
        """
        if isinstance(query, cls._item_class):
            return deepcopy(query)
        if isinstance(query, str):
            q = query.upper()
            for key, item in cls.catalog().items():
                if q == key.upper() or q in cls.get_item_aliases(item):
                    return deepcopy(item)
        msg = (
            f"Unknown {cls._item_class.__qualname__}: '{query}'. "
            f"Use one of {cls.every_aliases()}."
        )
        raise InvalidIcclimArgumentError(msg)

    @classmethod
    def lookup_no_error(cls, query: T | str) -> T | None:
        """
        Also look up an item in the registry, but return None if not found.

        Parameters
        ----------
        query : T or str
            The item to look up. It can be either an instance of the item class or a
            string.

        Returns
        -------
        T or None
            The found item, or None if not found.

        """
        try:
            return cls.lookup(query)
        except InvalidIcclimArgumentError:
            return None

    @classmethod
    def every_aliases(cls) -> list[T]:
        """
        Return a list of all aliases for items in the registry.

        Returns
        -------
        list[T]
            A list of all aliases for items in the registry.

        """
        return list(map(cls.get_item_aliases, list(cls.catalog().values())))

    @staticmethod
    def get_item_aliases(item: T) -> list[str]:
        """
        Get the aliases for the given item.

        Parameters
        ----------
        item : T
            The item to get aliases for.

        Returns
        -------
        list[str]
            A list of aliases for the item.

        Notes
        -----
        Should be overridden in subclasses.
        """
        return [item.name.upper()]

    @classmethod
    def catalog(cls) -> dict[str, T]:
        """
        Return a dictionary of all items in the registry.

        Returns
        -------
        dict[str, T]
            A dictionary containing all items in the registry, where the keys are the
            item names and the values are the item instances.

        """
        return {k: v for k, v in cls.__dict__.items() if isinstance(v, cls._item_class)}

    @classmethod
    def values(cls) -> list[T]:
        """
        Return a list of all items in the registry.

        Returns
        -------
        list[T]
            A list containing all items in the registry.

        """
        return [v for v in cls.__dict__.values() if isinstance(v, cls._item_class)]



"""Contain the IndexGroup class and the IndexGroupRegistry class."""

from __future__ import annotations

import dataclasses
from typing import Any

from icclim._core.model.registry import Registry


@dataclasses.dataclass(init=False)
class IndexGroup:
    """
    Class representing a group of climate indices.

    Parameters
    ----------
    name : str
        The name of the index group.
    values : list[IndexGroup] | None, optional
        The list of index groups contained within this group. Defaults to None.

    Attributes
    ----------
    name : str
        The name of the index group.
    values : list[IndexGroup]
        The list of index groups contained within this group.

    """

    name: str
    values: list[IndexGroup]

    def __init__(self, name: str, values: list[IndexGroup] | None = None) -> None:
        self.name = name
        if values is None:
            self.values = [self]
        else:
            self.values = values

    def get_indices(self) -> list[Any]:
        """
        Get the list of indices belonging to this group.

        Returns
        -------
        list[Any]
            The list of indices belonging to this group.

        Notes
        -----
        The list of indices is obtained by filtering the EcadIndexRegistry values.
        The others indices are not considered.
        """
        from icclim.ecad.registry import EcadIndexRegistry

        return list(
            filter(lambda i: i.group in self.values, EcadIndexRegistry.values()),
        )

    def __or__(self, right: IndexGroup) -> IndexGroup:
        """
        Compose 2 IndexGroups to create a new one.

        Parameters
        ----------
        right : IndexGroup
            The IndexGroup to be composed with.

        Returns
        -------
        IndexGroup
            The composed IndexGroup.

        """
        return IndexGroup(f"{self.name}_{right.name}", [self, right])

    def __eq__(self, other: object) -> bool:
        """
        Check if two IndexGroup objects are equal.

        Parameters
        ----------
        other : IndexGroup
            The IndexGroup to compare with.

        Returns
        -------
        bool
            True if the two IndexGroup objects are equal, False otherwise.

        """
        if not isinstance(other, IndexGroup):
            return False
        return other.name == self.name


class IndexGroupRegistry(Registry[IndexGroup]):
    """Registry for IndexGroup instances."""

    _item_class = IndexGroup

    TEMPERATURE = IndexGroup("temperature")
    HEAT = IndexGroup("heat")
    COLD = IndexGroup("cold")
    DROUGHT = IndexGroup("drought")
    RAIN = IndexGroup("rain")
    SNOW = IndexGroup("snow")
    WIND = IndexGroup("wind")
    PRESSURE = IndexGroup("pressure")
    SUNSHINE = IndexGroup("sunshine")
    HUMIDITY = IndexGroup("humidity")
    # no climate index should be bounded to "all"
    WILD_CARD_GROUP = IndexGroup("all")



"""Contain the UserIndexDict TypedDict."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Sequence

    from icclim._core.legacy.user_index.calc_operation import (
        CalcOperation,
        CalcOperationLike,
    )
    from icclim._core.model.logical_link import LogicalLink


class UserIndexDict(TypedDict, total=False):
    """
    User index dictionary.

    This dictionary is used to describe how a user index should be setup to compute a
    custom climate index.
    The reccomended way to create a custom indices is now to use the
    generic indices API.
    See :ref:`generic_indices_recipes` for how to combine thresholds with generic
    """

    index_name: str
    calc_operation: CalcOperationLike | CalcOperation
    logical_operation: str | None | Sequence[str]  # >= | <= | ...| ==
    thresh: str | int | Sequence[str | int] | None
    extreme_mode: Literal["min", "max"] | None

    link_logical_operations: Literal["and", "or"] | LogicalLink | None
    coef: float | None
    date_event: bool
    var_type: Literal["t", "p"] | None
    window_width: int | None
    ref_time_range: Sequence[dt.datetime] | Sequence[str] | tuple[str, str] | None

    # -- deprecated
    indice_name: str | None



"""Factory to build a `Threshold` from a query or from its components."""

from __future__ import annotations

import re
from collections.abc import Sequence
from typing import TYPE_CHECKING

from pint.errors import UndefinedUnitError
from xarray import DataArray, Dataset
from xclim.core.units import units as xc_units

from icclim._core.constants import (
    DOY_PERCENTILE_UNIT,
    PERIOD_PERCENTILE_UNIT,
)
from icclim._core.generic.threshold.basic import BasicThreshold
from icclim._core.generic.threshold.bounded import BoundedThreshold
from icclim._core.generic.threshold.percentile import PercentileThreshold
from icclim._core.input_parsing import (
    PercentileDataArray,
    find_standard_vars,
    get_name_of_first_var,
    is_dataset_path,
    read_dataset,
)
from icclim._core.model.logical_link import LogicalLink, LogicalLinkRegistry
from icclim._core.model.operator import Operator, OperatorRegistry
from icclim._core.model.threshold import (
    Threshold,
    ThresholdBuilderInput,
    ThresholdValueType,
)
from icclim._core.utils import is_number_sequence
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    import pint

VALUE_REGEX = re.compile(r"-?\d+\.?\d*")
OPERAND_REGEX = re.compile(r"[<>=]")


def build_threshold(
    query: str | None = None,
    *,
    operator: Operator | str | None = None,
    value: ThresholdValueType = None,
    unit: str | None = None,
    threshold_min_value: str | float | pint.Quantity | None = None,
    thresholds: Sequence[Threshold | str] | None = None,
    logical_link: str | LogicalLink | None = None,
    offset: str | float | pint.Quantity | None = None,
    **kwargs,
) -> BoundedThreshold | PercentileThreshold | BasicThreshold:
    """
    Build a `Threshold`.

    This function is a factory for `Threshold` instances.
    It can build a `BasicThreshold`, a `PercentileThreshold` or a `BoundedThreshold`.
    See :ref:`generic_indices_recipes` for how to combine thresholds with generic
    indices.

    Parameters
    ----------
    query: str | None = None
        string query describing a threshold.
        It must include: an operator, a threshold value and optionally a unit
        such as "> 10 degC".
        It acts as a shorthand for ``operator``, ``value`` and ``unit`` parameters for
        simple threshold.
        Don't use ``query`` when value is a DataArray, a Dataset or a path to a
        netcdf/zarr storage, instead use ``operator``, ``value`` and ``unit``.
        ``query`` supersede `operator`, `value` and `unit` parameters.
    operator: Operator | str = None
        keyword argument only.
        The operator either as an instance of Operator or as a compatible string.
        See :py:class:`OperatorRegistry` for the list of all operators.
        When ``query`` is None and operator is None, the default ``Operator.REACH`` is
        used.
    value: str | float | int | Dataset | DataArray | Sequence[float | int | str] | None
        keyword argument only.
        The threshold value(s), default to None.
        It can be:
        * a simple scalar threshold
        * a percentile that will be computed per-grid cell (in combinaison with `unit`)
        * per-grid cell thresholds defined by a DataArray, a Dataset or a string path to
        a netcdf/zarr.
        * a sequence of scalars, the indicator will then be computed for each value and
        a specific dimension will be created (also work with percentiles).
    unit: str | None = None
        Keyword argument only.
        The threshold unit.
        When ``unit`` is None, if ``value`` is a dataset and a "units"
        can be read from its `attrs`, this unit will be used. If value is a scalar or
        a sequence of scalar, the exceedance will be computed by assuming threshold has
        the same unit as the studied value is it compared to.
        When unit is a string it must be a valid unit of our shared pint registry with
        xclim or a special percentile unit:
        * "doy_per" for day of year percentiles (in ECAD, used for temperature indices
        such as TX90p)
        * "period_per" for per period percentiles (in ECAD, used for rain indices such
        as R75p)
    threshold_min_value: str | float | pint.Quantity
        A minimum value used to pre-filter computed threshold values.
        This is particularly useful to compute a percentile threshold for rain where
        dry days are usually ignored. In that case threshold_min_value would be set to
        "1 mm/day".
        If threshold_min_value is a number, ``unit`` is used to quantify
        ``threshold_min_value``.
    offset: float | None
        Optional. An offset applied to the threshold. This is particularly useful when
        the threshold is an existing dataset (netcdf file or zarr store) and the data
        should be compared to this dataset + an offset
        (e.g. to compute days when T > 5 degC above normal)
    kwargs
        Additional arguments to build a PercentileThreshold.
        See :py:class:`PercentileThreshold` constructor for the complete list
        of possible arguments.

    Examples
    --------
    .. code-block:: python

        # -- Scalar threshold
        scalar_t = build_threshold(">= 30 degC")
        assert isinstance(scalar_t, BasicThreshold)

        # -- Daily percentile threshold
        doy_t = build_threshold(">= 30 doy_per")
        assert isinstance(doy_t, PercentileThreshold)

        # -- Per grid-cell threshold, with offset
        grided_t = build_threshold(
            operator=">=", value="path/to/tasmax_thresholds.nc", unit="K", offset=5
        )
        assert isinstance(grided_t, BasicThreshold)

        # -- Daily percentile threshold, from a file
        tasmax = xarray.open_dataset("path/to/tasmax_thresholds.nc").tasmax
        doys = xclim.core.calendar.percentile_doy(tasmax)
        doy_file_t = build_threshold(operator=">=", value=doys)
        assert isinstance(doy_file_t, PercentileThreshold)

        # -- Bounded threshold
        bounded_t = build_threshold(">= -20 degree AND <= 20 degree ")
        # equivalent to:
        x = build_threshold(">= -20 degree")
        y = build_threshold("<= 20 degree")
        bounded_t2 = x & y
        assert bounded_t == bounded_t2
        # equivalent to:
        bounded_t3 = build_threshold(thresholds=[x, y], logical_link="AND")
        assert bounded_t == bounded_t3
        assert isinstance(bounded_t, BoundedThreshold)

    """
    input_thresh = _read_input(
        query,
        operator,
        value,
        unit,
        threshold_min_value,
        thresholds,
        logical_link,
        offset,
        **kwargs,
    )
    if _must_build_per_threshold(input_thresh):
        return PercentileThreshold(**input_thresh)
    if _must_build_basic_threshold(input_thresh):
        return BasicThreshold(**input_thresh)
    if _must_build_bounded_threshold(input_thresh):
        return BoundedThreshold(**input_thresh)
    if _must_build_per_grid_cell_threshold(input_thresh):
        return BasicThreshold(**input_thresh)
    msg = f"Threshold cannot be built from a {type(value)}"
    raise NotImplementedError(msg)


def _get_operator(query: str) -> tuple[Operator | None, str]:
    operand = "".join(OPERAND_REGEX.findall(query))
    op = OperatorRegistry.lookup_no_error(operand)
    if op is not None:
        return op, query.replace(op.operand, "", 1)
    return None, query


def _read_string_threshold(query: str) -> tuple[str, str, float]:
    op, no_op_query = _get_operator(query)
    operand = op.operand if op else ""
    if DOY_PERCENTILE_UNIT in no_op_query:
        val = re.findall(VALUE_REGEX, no_op_query)[0]
        unit = DOY_PERCENTILE_UNIT
    elif PERIOD_PERCENTILE_UNIT in no_op_query:
        val = re.findall(VALUE_REGEX, no_op_query)[0]
        unit = PERIOD_PERCENTILE_UNIT
    else:
        try:
            quantity = xc_units.Quantity(no_op_query)
        except UndefinedUnitError as e:
            msg = f"Could not build threshold from {query}"
            raise InvalidIcclimArgumentError(msg) from e
        val = quantity.m
        unit = None if quantity.unitless else str(quantity.units)
    return operand, unit, val


def _build_quantity(
    quantity: None | str | float | pint.Quantity,
    default_unit: str | None,
) -> pint.Quantity | None:
    if quantity is None:
        return None
    if isinstance(quantity, xc_units.Quantity):
        return quantity
    if isinstance(quantity, (float, int)):
        if default_unit in (PERIOD_PERCENTILE_UNIT, DOY_PERCENTILE_UNIT):
            unit = None
        else:
            unit = default_unit
        return xc_units.Quantity(value=quantity, units=unit)
    if isinstance(quantity, str):
        operator, unit, value = _read_string_threshold(quantity)
        if operator is not None and operator != "":
            msg = (
                f"Cannot parse quantity '{quantity}'"
                f" The operator {operator} should be removed."
            )
            raise InvalidIcclimArgumentError(msg)
        return xc_units.Quantity(value=value, units=unit)
    msg = f"Unknown type '{type(quantity)}' for quantity {quantity}"
    raise NotImplementedError(msg)


def _read_input(
    query: str | None = None,
    operator: Operator | str | None = None,
    value: ThresholdValueType = None,
    unit: str | None = None,
    threshold_min_value: str | float | None = None,
    thresholds: tuple[Threshold, Threshold] | None = None,
    logical_link: str | LogicalLink | None = None,
    offset: str | float | pint.Quantity | None = None,
    **kwargs,
) -> ThresholdBuilderInput:
    if query is not None and _must_read_query(query, operator, value, unit):
        if _is_bounded_threshold_query(query):
            return _read_bounded_threshold_query(query)
        return _read_threshold_from_query(query, threshold_min_value, kwargs)
    if _must_read_bounded(operator, value, unit, thresholds, logical_link):
        return _read_bounded_threshold(thresholds, logical_link)
    if operator is not None:
        if (operator := OperatorRegistry.lookup_no_error(operator)) is None:
            operator = OperatorRegistry.REACH
        return {
            "operator": operator,
            "unit": unit,
            "value": value,
            "threshold_min_value": _build_quantity(threshold_min_value, unit),
            "offset": _build_quantity(offset, unit),
            **kwargs,
        }
    msg = "Could not read threshold"
    raise NotImplementedError(msg)


def _read_bounded_threshold(
    thresholds: tuple[Threshold, Threshold],
    logical_link: LogicalLink | str,
) -> ThresholdBuilderInput:
    acc = []
    for t in thresholds:
        if isinstance(t, str):
            acc.append(_read_input(t))
        elif isinstance(t, Threshold):
            acc.append(t)
        else:
            msg = f"Unknown type '{type(t)}'"
            raise NotImplementedError(msg)
    if len(acc) > 2:
        msg = "Can't build BoundedThreshold on more than 2 thresholds."
        raise NotImplementedError(msg)
    if isinstance(logical_link, str):
        logical_link = LogicalLinkRegistry.lookup(logical_link)
    return {
        "initial_query": None,
        "thresholds": tuple(acc),
        "logical_link": logical_link,
    }


def _read_threshold_from_query(
    query: str,
    threshold_min_value: None | str | float | pint.Quantity,
    kwargs: dict,
) -> ThresholdBuilderInput:
    operator, unit, value = _read_string_threshold(query)
    if (operator := OperatorRegistry.lookup_no_error(operator)) is None:
        operator = OperatorRegistry.REACH
    return {
        "operator": operator,
        "unit": unit,
        "value": value,
        "threshold_min_value": _build_quantity(threshold_min_value, unit),
        "initial_query": query,
        **kwargs,
    }


def _must_read_query(
    query: str,
    operator: Operator | str | None,
    value: ThresholdValueType | None,
    unit: str | None,
) -> bool:
    return (
        isinstance(query, str) and operator is None and value is None and unit is None
    )


def _must_read_bounded(
    operator: Operator | str | None,
    value: ThresholdValueType,
    unit: str | None,
    thresholds: tuple[Threshold, Threshold] | None,
    logical_link: str | LogicalLink | None,
) -> bool:
    return (
        operator is None
        and value is None
        and unit is None
        and thresholds is not None
        and logical_link is not None
    )


def _is_bounded_threshold_query(query: str) -> bool:
    return any(
        l_l.name.upper() in query.upper() for l_l in LogicalLinkRegistry.values()
    )


def _read_bounded_threshold_query(query: str) -> ThresholdBuilderInput:
    link = None
    split_word = None
    uppered = query.upper()
    for l_l in LogicalLinkRegistry.values():
        if l_l.name.upper() in uppered:
            link = l_l
            index_of_link = uppered.index(l_l.name.upper())
            split_word = query[index_of_link : index_of_link + len(l_l.name)]
            break
    if link is None:
        msg = f"No logical link found in {query}"
        raise InvalidIcclimArgumentError(msg)
    threshs = [t.strip() for t in query.split(split_word)]
    if len(threshs) != 2:
        msg = (
            "BoundedThreshold can only be built on 2"
            f" thresholds. We found {len(threshs)} here."
        )
        raise InvalidIcclimArgumentError(msg)
    return {
        "initial_query": query,
        "thresholds": (_read_input(threshs[0]), _read_input(threshs[1])),
        "logical_link": link,
    }


def _must_build_per_threshold(builder_input: ThresholdBuilderInput) -> bool:
    value = builder_input.get("value")
    unit = builder_input.get("unit", None)
    var_name = builder_input.get("threshold_var_name", None)
    return _has_per_unit(unit, value) or _is_per_dataset(var_name, value)


def _is_per_dataset(threshold_var_name: str, value: str | Dataset | DataArray) -> bool:
    if is_dataset_path(value) or isinstance(value, Dataset):
        thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
    else:
        thresh_da = value
    return PercentileDataArray.is_compatible(thresh_da)


def _has_per_unit(unit: str | None | Sequence[float], value: float) -> bool:
    return isinstance(value, (float, Sequence)) and (
        unit in (DOY_PERCENTILE_UNIT, PERIOD_PERCENTILE_UNIT)
    )


def _must_build_per_grid_cell_threshold(builder_input: ThresholdBuilderInput) -> bool:
    value = builder_input.get("value")
    threshold_var_name = builder_input.get("threshold_var_name", None)
    if value is None:
        # Case of comparison to normal like dcsc::TxND,
        # the threshold is setup but ::prepare must be called
        # to initialize the value
        return True
    if is_dataset_path(value) or isinstance(value, Dataset):
        thresh_da = _get_dataarray_from_dataset(threshold_var_name, value)
        return not PercentileDataArray.is_compatible(thresh_da)
    return False


def _must_build_basic_threshold(builder_input: ThresholdBuilderInput) -> bool:
    value = builder_input.get("value")
    return is_number_sequence(value) or isinstance(value, (DataArray, float, int))


def _must_build_bounded_threshold(builder_input: ThresholdBuilderInput) -> bool:
    logical_link = builder_input.get("logical_link")
    return logical_link is not None


def _get_dataarray_from_dataset(
    threshold_var_name: str | None,
    value: Dataset | str,
) -> DataArray:
    ds = value if isinstance(value, Dataset) else read_dataset(value, standard_var=None)
    if threshold_var_name is None:
        if len(ds.data_vars) == 1:
            threshold_var_name = get_name_of_first_var(ds)
        else:
            names = find_standard_vars(ds)
            if len(names) == 1:
                threshold_var_name = names[0]
            else:
                msg = (
                    f"Could not guess the variable to use as a threshold in {ds}."
                    f" Use `threshold_var_name` to specify which variable should be"
                    f" used."
                )
                raise InvalidIcclimArgumentError(msg)
    return ds[threshold_var_name]



"""BasicThreshold module."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import numpy as np
import xarray as xr
from xarray import DataArray, Dataset
from xclim.core.units import convert_units_to, str2pint

from icclim._core.constants import (
    UNITS_KEY,
)
from icclim._core.generic.threshold.threshold_templates import (
    EN_THRESHOLD_TEMPLATE,
    ThresholdMetadata,
)
from icclim._core.input_parsing import (
    get_dataarray_from_dataset,
    is_dataset_path,
)
from icclim._core.model.threshold import (
    Threshold,
    ThresholdValueType,
)
from icclim._core.utils import is_number_sequence
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    import jinja2
    import pint

    from icclim._core.model.operator import Operator


class BasicThreshold(Threshold):
    """
    Pint ready simple threshold (e.g. "> 300 K").

    Parameters
    ----------
    operator : Operator or str
        The operator used for the threshold comparison.
    value : ThresholdValueType
        The threshold value(s) to compare against.
    unit : str, optional
        The unit of the threshold value(s).
    initial_query : str, optional
        The initial query used to build the threshold.
    threshold_min_value : pint.Quantity, optional
        The minimum value for the threshold.
    threshold_var_name : str, optional
        The name of the threshold variable.
    offset : pint.Quantity, optional
        The offset to be applied to the threshold value(s).

    Notes
    -----
    When built, `value` is always turned into a `xarray.DataArray`.
    The `unit` property has a setter that will attempt a unit conversion using
    units found in xclim's pint registry.

    The actual unit can be overridden by modifying `value.attrs["units"]` directly.
    """

    @property
    def unit(self) -> str | None:
        """The unit of the threshold value(s)."""
        if self.value is not None:
            return self.value.attrs[UNITS_KEY]
        return self._unit

    @unit.setter
    def unit(self, unit: str) -> None:
        """
        Set the unit of the threshold value(s).

        Parameters
        ----------
        unit : str
            The unit to set.

        Notes
        -----
        This setter will attempt a unit conversion using units found in xclim's pint
        registry.
        """
        if self.value is not None:
            if self.value.attrs.get(UNITS_KEY, None) is not None and unit is not None:
                self.value = convert_units_to(self.value, unit, context="hydro")
            self.value.attrs[UNITS_KEY] = unit
        else:
            self._unit = unit

    def __init__(
        self,
        operator: Operator | str,
        value: ThresholdValueType,
        unit: str | None = None,
        initial_query: str | None = None,
        threshold_min_value: pint.Quantity | None = None,
        threshold_var_name: str | None = None,
        offset: pint.Quantity | None = None,
        **kwargs,  # noqa: ARG002
    ) -> None:
        """
        Initialize a Threshold object.

        Parameters
        ----------
        operator : Operator or str
            The operator used for comparison. Can be an Operator object or a string
            representation of the operator.
        value : ThresholdValueType
            The threshold value. Can be a scalar, a sequence of numbers, a Dataset, or a
            DataArray.
        unit : str or None, optional
            The unit of the threshold value. Defaults to None.
        initial_query : str or None, optional
            The initial query string. Defaults to None.
        threshold_min_value : pint.Quantity or None, optional
            The minimum threshold value. Defaults to None.
        threshold_var_name : str or None, optional
            The name of the threshold variable. Defaults to None.
        offset : pint.Quantity or None, optional
            The offset value. Defaults to None.
        **kwargs
            Additional keyword arguments.

        Raises
        ------
        InvalidIcclimArgumentError
            If threshold_min_value is used with scalar thresholds.
        NotImplementedError
            If the threshold value type is not supported.

        """
        if (
            is_number_sequence(value) or isinstance(value, (float, int))
        ) and threshold_min_value is not None:
            msg = "Cannot use threshold_min_value with scalar thresholds."
            raise InvalidIcclimArgumentError(msg)
        if is_dataset_path(value) or isinstance(value, Dataset):
            # e.g. build_threshold(">", "thresh*.nc" , "degC") noqa: ERA001
            thresh_da = get_dataarray_from_dataset(threshold_var_name, value)
            self.value = self._prepare_da(thresh_da, threshold_min_value, offset, unit)
            unit = self.value.attrs.get(UNITS_KEY, None)
        elif isinstance(value, DataArray):
            self.value = self._prepare_da(value, threshold_min_value, offset, unit)
            unit = self.value.attrs.get(UNITS_KEY, None)
        elif is_number_sequence(value):
            self.value = DataArray(
                name="threshold",
                data=value,
                attrs={UNITS_KEY: unit},
                dims="threshold",
                coords={"threshold": value},
            )
        elif isinstance(value, (float, int)):
            # e.g. build_threshold(">", [2,3,4], "degC") noqa: ERA001
            self.value = DataArray(
                name="threshold",
                data=value,
                attrs={UNITS_KEY: unit},
            )
        elif value is None:
            self.prepare = self._partial_prepare_da(threshold_min_value, offset, unit)
            self.is_ready = False
            self.value = None
        else:
            msg = f"Cannot build threshold from a {type(value)}."
            raise NotImplementedError(msg)
        self.operator = operator
        self.unit = unit
        self.is_ready = True
        self.threshold_var_name = threshold_var_name
        self.initial_query = initial_query
        self.offset = offset
        if threshold_min_value is not None and threshold_min_value.dimensionless:
            self.threshold_min_value = float(threshold_min_value.m) * str2pint(unit)
        else:
            self.threshold_min_value = threshold_min_value

    def __eq__(self, other: object) -> bool:
        """Check if two BasicThreshold objects are equal."""
        return (
            isinstance(other, BasicThreshold)
            and self.operator == other.operator
            and self.value == other.value
            and self.unit == other.unit
            and self.initial_query == other.initial_query
            and self.threshold_min_value == other.threshold_min_value
        )

    def __str__(self) -> str:
        """Return a string representation of the threshold."""
        return f"""Threshold:
        Operator: {self.operator}
        Value: {self.value}
        Unit: {self.unit}
        Initial Query: {self.initial_query}
        Threshold Min Value: {self.threshold_min_value}
        Offset: {self.offset}
        """

    def format_metadata(
        self,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
        **kwargs,  # noqa: ARG002
    ) -> ThresholdMetadata:
        """
        Generate the metadata for the threshold.

        These metadata are used to fill the generic template.

        Parameters
        ----------
        jinja_scope : dict
            The jinja scope, it contains the variables to be used in the jinja template.
        jinja_env : jinja2.Environment
            The jinja environment, it contains the jinja rendering engine.
        **kwargs
            Additional keyword arguments, ignored for compatibility with other
            `format_metadata` methods.

        Returns
        -------
        ThresholdMetadata
            The metadata for the threshold.

        """
        templates = self._get_metadata_templates()
        conf = {
            "operator": self.operator,
            "unit": self.unit,
            "threshold_min_value": self.threshold_min_value,
            "has_offset": self.offset is not None,
            "offset": self.offset,
            "must_run_bootstrap": True,
            "value": self.value,
        }
        conf.update(jinja_scope)
        return {
            k: jinja_env.from_string(v, globals=conf).render()
            for k, v in templates.items()
        }

    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,  # noqa: ARG002
    ) -> DataArray:
        """
        Compute the threshold exceedance value.

        Parameters
        ----------
        comparison_data : xr.DataArray
            The data array to compare with the threshold value.
        override_op : Callable[[DataArray, DataArray], DataArray] | None, optional
            A custom override function to compute the threshold value.
            If provided, this function will be used instead of the default operator.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        DataArray
            The computed threshold value.

        Notes
        -----
        If `override_op` is not None, the `override_op` function will be used to
        compute the threshold exceedance using the `comparison_data` and `self.value`
        as inputs.
        If `override_op` is None, the default operator defined in `self.operator`
        will be used to compute the threshold exceedance.

        """
        if override_op is not None:
            return override_op(comparison_data, self.value)
        return self.operator.compute(comparison_data, self.value)

    def _partial_prepare_da(
        self,
        min_val: pint.Quantity | None,
        offset: pint.Quantity | None,
        unit: str | None,
    ) -> Callable[[xr.DataArray], xr.DataArray]:
        def _final_prepare_da(value: xr.DataArray) -> xr.DataArray:
            result = self._prepare_da(value, min_val, offset, unit)
            self.value = result
            return result

        return _final_prepare_da

    def _get_metadata_templates(self) -> ThresholdMetadata:
        if self.value.size == 1:
            return EN_THRESHOLD_TEMPLATE["single_value"]
        return EN_THRESHOLD_TEMPLATE["multiple_values"]

    def _prepare_da(
        self,
        value: xr.DataArray,
        min_value: pint.Quantity | None,
        offset: pint.Quantity | None,
        unit: str | None,
    ) -> xr.DataArray:
        built_value = _apply_min_value(value, min_value)
        built_value = _apply_offset(built_value, offset)
        if unit is not None:
            built_value = convert_units_to(built_value, unit, context="hydro")
        self.is_ready = True
        return built_value


def _apply_offset(da: DataArray, offset: pint.Quantity | None) -> xr.DataArray:
    if offset is not None:
        if offset.dimensionless:
            # We assume the same unit as `da` if it's dimensionless
            offset = offset.m
        else:
            offset = convert_units_to(str(offset), da, context="hydro")
        with xr.set_options(keep_attrs=True):
            return da + offset
    return da


def _apply_min_value(da: DataArray, min_value: pint.Quantity | None) -> xr.DataArray:
    if min_value is not None:
        if min_value.dimensionless:
            # We assume min_value use the same unit as thresh_da if it's dimensionless
            min_value = min_value.m
        else:
            min_value = convert_units_to(str(min_value), da, context="hydro")
        return da.where(da > min_value, np.nan)
    return da



"""European Climate Assesment & Dataset (ECAD) indices.

The ECAD indices public API, via `icclim.ecad` package, is generated from the
`icclim.ecad.registry.EcadIndexRegistry` registry definitions.
The parameters of the functions are specialized to each index but are all taken from
`icclim.main.index` general function.
In other words, the ECAD indices in `icclim.ecad` package are specializations
of `icclim.main.index` for ECAD indices.
"""

from icclim._generated._ecad import *  # noqa: F403



"""Type hints for icclim."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal, Union

from xarray import DataArray, Dataset

from icclim._core.model.in_file_dictionary import InFileDictionary

InFileBaseType = Union[str, Sequence[str], Dataset, DataArray]
ThresholdedDict = dict[str, InFileDictionary]
InFileLike = Union[ThresholdedDict, InFileBaseType, dict[str, InFileBaseType]]

FrequencyLike = Union[str, list[str | tuple | int], tuple[str, list | tuple]]
# MonthsIndexer format: [12,1,2,3]
MonthsIndexer = dict[Literal["month"], Sequence[int]]
# DatesIndexer format: ("01-25", "02-28")
DatesIndexer = dict[Literal["date_bounds"], tuple[str, str]]
Indexer = Union[MonthsIndexer, DatesIndexer]

SamplingMethodLike = Literal["groupby", "resample", "groupby_ref_and_resample_study"]

ThresholdValueType = Union[
    str,
    float,
    int,
    Dataset,
    DataArray,
    Sequence[float | int | str],
    None,
]



"""Generic indices.

The generic indices public API, via `icclim.generic` package, is generated from the
`icclim.generic.registry.GenericIndicatorRegistry` registry definitions.
icclim's generic indices are a generalization of the climate indices found in
ECAD and DCSC's registries.
They can be computed on any dataset and make use of the `Threshold` interface
to enable the creation of personalized indices.
The parameters of the functions are specialized to each index but are all taken from
`icclim.main.index` general function.
In other words, the generic indices in `icclim.generic` package are specializations
of `icclim.main.index` for ECAD indices.

Examples
--------
>>> from icclim.generic import count_occurrences
>>> from icclim import build_threshold
>>> thresh = build_threshold(">= 25 °C and <= 30 °C")
>>> result = count_occurrences("tas.nc", thresh).compute()
>>> print(result.count_occurrences)

"""

# from icclim._generated._generic import *



"""Module for the Operator class and OperatorRegistry."""

from __future__ import annotations

import dataclasses
import operator
from collections.abc import Callable
from typing import TYPE_CHECKING, NoReturn

from icclim._core.model.registry import Registry
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    from xarray import DataArray


def _reach_err(_: object, __: object) -> NoReturn:
    # can't raise error in lambda
    msg = (
        "Reach operator can't be called. Try to fill threshold with an operand"
        " (e.g. '>=' in '>= 22 degC')."
    )
    raise InvalidIcclimArgumentError(msg)


@dataclasses.dataclass
class Operator:
    """
    Represents an operator used in computations.

    Parameters
    ----------
    short_name : str
        The short name of the operator,
        used when templating the output metadata.
    long_name : str
        The long name of the operator,
        used when templating the output metadata.
    standard_name : str
        The standard name of the operator,
        used when templating the output metadata.
    operand : str
        The operand symbol of the operator.
    compute : Callable[[DataArray, DataArray | int | float], DataArray]
        The computation function of the operator.
    aliases : list[str]
        The list of aliases for the operator.
    """

    short_name: str
    long_name: str
    standard_name: str
    operand: str
    compute: Callable[[DataArray, DataArray | int | float], DataArray]
    aliases: list[str]

    def __call__(self, *args, **kwargs) -> DataArray:
        """
        Call the computation function of the operator.

        Parameters
        ----------
        *args : tuple
            Positional arguments passed to the computation function.
        **kwargs : dict
            Keyword arguments passed to the computation function.

        Returns
        -------
        DataArray
            The result of the computation.

        """
        return self.compute(*args, **kwargs)


class OperatorRegistry(Registry[Operator]):
    """
    Registry of operators.

    Contains the predefined operators used to build ``Threshold``.
    """

    _item_class = Operator

    @staticmethod
    def get_item_aliases(op: Operator) -> list[str]:
        """
        Get the aliases of an operator.

        Parameters
        ----------
        op : Operator
            The operator.

        Returns
        -------
        list[str]
            The list of aliases for the operator.
        """
        return list(map(str.upper, op.aliases))

    GREATER = Operator(
        short_name="gt",
        long_name="greater than",
        standard_name="greater_than",
        aliases=["gt", ">"],
        operand=">",
        compute=operator.gt,
    )
    LOWER = Operator(
        short_name="lt",
        long_name="lower than",
        standard_name="lower_than",
        aliases=["lt", "<"],
        operand="<",
        compute=operator.lt,
    )
    GREATER_OR_EQUAL = Operator(
        short_name="get",
        long_name="greater or equal to",
        standard_name="greater_or_equal_to",
        aliases=["get", "ge", ">=", "=>"],
        operand=">=",
        compute=operator.ge,
    )
    LOWER_OR_EQUAL = Operator(
        short_name="let",
        long_name="lower or equal to",
        standard_name="lower_or_equal_to",
        aliases=["let", "le", "<=", "=<"],
        operand="<=",
        compute=operator.le,
    )
    EQUAL = Operator(
        short_name="e",
        long_name="equal to",
        standard_name="equal_to",
        aliases=["e", "equal", "eq", "=", "=="],
        operand="==",
        compute=operator.eq,
    )
    # A None operand means the threshold is reached and a reducer specific computation
    # is done. Case of excess and deficit (a.k.a gd4, hd17)
    REACH = Operator(
        short_name="reach",
        long_name="",  # nothing
        standard_name="reaching",
        aliases=["r"],
        operand="reach",
        compute=_reach_err,
    )



"""Contain the LogicalLink class and registry."""

from __future__ import annotations

import dataclasses
from collections.abc import Callable
from functools import reduce
from typing import TYPE_CHECKING

import numpy as np

from icclim._core.model.registry import Registry

if TYPE_CHECKING:
    from xarray import DataArray


@dataclasses.dataclass
class LogicalLink:
    """Logical link class to combine multiple threshold.

    This is meant to be used with the old user_indices API.
    It is now reccomended to use BoundedThreshold with generic indices instead.
    See :ref:`generic_indices_recipes` for how to combine thresholds with generic
    indices.
    """

    name: str
    standard_name: str
    long_name: str
    short_name: str
    compute: Callable[[list[DataArray]], DataArray]

    def __call__(self, *args, **kwargs) -> DataArray:
        """Compute the logical "and" or "or" of the input data."""
        return self.compute(*args, **kwargs)


class LogicalLinkRegistry(Registry[LogicalLink]):
    """Registry for LogicalLink objects."""

    _item_class = LogicalLink

    LOGICAL_OR = LogicalLink(
        name="or",
        standard_name="OR",
        long_name="OR",
        short_name="OR",
        compute=lambda data_list: reduce(np.logical_or, data_list),
    )
    LOGICAL_AND = LogicalLink(
        name="and",
        standard_name="AND",
        long_name="AND",
        short_name="AND",
        compute=lambda data_list: reduce(np.logical_and, data_list),
    )
